import numpy as np
cimport numpy as np

cdef struct Relation:
    np.double_t reciprocity
    # np.double_t upper_bound
    # np.double_t jump
    np.intp_t target
    # np.intp_t my_rank
    # np.intp_t rec_rank
    # np.intp_t my_scope
    # np.double_t my_dis
    np.double_t dia_dis
    # np.intp_t penalty
    # np.intp_t my_members
    # np.intp_t rec_members
    # np.intp_t loop_rank
    # np.intp_t index
    # np.double_t value



# cython: language_level=3
# cython: boundscheck=False
# cython: nonecheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True

# Builds minimum spanning tree for druhg algorithm
# uses dialectics to evaluate reciprocity
# Author: Pavel "DRUHG" Artamonov
# License: 3-clause BSD

import numpy as np
cimport numpy as np
import sys

import _heapq as heapq
from libc.math cimport fabs, pow
import bisect

cdef np.double_t INF = sys.float_info.max

from sklearn.neighbors import KDTree, BallTree
# from sklearn import preprocessing
from joblib import Parallel, delayed

cdef class PairwiseDistanceTreeSparse(object):
    cdef object data_arr
    cdef int data_size

    def __init__(self, N, d):
        self.data_size = N
        self.data_arr = d

    cpdef tuple query(self, d, k, dualtree = 0, breadth_first = 0):
        # TODO: actually we need to consider replacing INF with something else.
        # Reciprocity of absent link is not the same as the INF. Do reciprocity with graphs!
        cdef np.ndarray[np.double_t, ndim=2] knn_dist
        cdef np.ndarray[np.intp_t, ndim=2] knn_indices

        knn_dist = INF*np.ones((self.data_size, k+1))
        knn_indices = np.zeros((self.data_size, k+1), dtype=np.intp)

        warning = 0

        i = self.data_size
        while i:
            i -= 1
            row = self.data_arr.getrow(i)
            idx, data = row.indices, row.data
            sorted = np.argsort(data)
            j = min(k,len(idx))
            if idx[sorted[0]] != i:
                while j:
                    j -= 1
                    knn_dist[i][j+1] = data[sorted[j]]
                    knn_indices[i][j+1] = idx[sorted[j]]
            else:
                # edge loops itself
                warning += 1
                while j:
                    j -= 1
                    knn_dist[i][j] = data[sorted[j]]
                    knn_indices[i][j] = idx[sorted[j]]

            knn_dist[i][0], knn_indices[i][0] = 0., i # have to add itself. Edge to itself have to be zero!

        if warning:
            print ('Attention!: Sparse matrix has an edge that forms a loop! They were zeroed.', warning)

        return knn_dist, knn_indices

cdef class PairwiseDistanceTreeGeneric(object):
    cdef object data_arr
    cdef int data_size

    def __init__(self, N, d):
        self.data_size = N
        self.data_arr = d

    cpdef tuple query(self, d, k, dualtree = 0, breadth_first = 0):
        cdef np.ndarray[np.double_t, ndim=2] knn_dist
        cdef np.ndarray[np.intp_t, ndim=2] knn_indices

        knn_dist = np.zeros((self.data_size, k))
        knn_indices = np.zeros((self.data_size, k), dtype=np.intp)

        i = self.data_size
        while i:
            i -= 1
            row = self.data_arr[i]
            sorted = np.argsort(row)
            j = k
            while j:
                j -= 1
                knn_dist[i][j] = row[sorted[j]]
                knn_indices[i][j] = sorted[j]

        return knn_dist, knn_indices

cdef class UnionFind (object):
    cdef:
        np.ndarray parent_arr
        np.intp_t *parent

        np.intp_t next_label

    def __init__(self, np.intp_t N):
        self.parent_arr = np.zeros(2 * N, dtype=np.intp)
        self.parent = (<np.intp_t *> self.parent_arr.data)
        self.next_label = N + 1

    cdef np.intp_t fast_find(self, np.intp_t n):
        cdef np.intp_t p, temp

        p = self.parent[n]
        if p == 0:
            return n
        while self.parent[p] != 0:
            p = self.parent[p]

        # label up to the root
        while p != n:
            temp = self.parent[n]
            self.parent[n] = p
            n = temp

        return p

    # cdef np.intp_t is_cluster(self, np.intp_t n):
    #     return self.parent[n]

    cdef void union(self, np.intp_t aa, np.intp_t bb):
        aa, bb = self.fast_find(aa), self.fast_find(bb)

        self.parent[aa] = self.parent[bb] = self.next_label
        self.next_label += 1
        return

cdef class UniversalReciprocity (object):
    """Constructs DRUHG spanning tree and marks parents of clusters

    Parameters
    ----------

    algorithm : int
        0/1 - for KDTree/BallTree object
        2/3 - for a full/scipy.sparse precomputed pairwise squared distance matrix

    data: object
        Pass KDTree/BallTree objects or pairwise matrix.

    max_neighbors_search : int, optional (default= 16)
        The max_neighbors_search parameter of DRUHG.
        Effects performance vs precision.
        Default is more than enough.

    metric : string, optional (default='euclidean')
        The metric used to compute distances for the tree.
        Used only with KDTree/BallTree option.

    leaf_size : int, optional (default=20)
        sklearn K-NearestNeighbor uses it.
        Used only with KDTree/BallTree option.

    **kwargs :
        Keyword args passed to the metric.
        Used only with KDTree/BallTree option.
    """

    cdef:
        object tree
        object dist_tree

        np.double_t PRECISION

        np.intp_t num_points
        np.intp_t num_features

        np.intp_t max_neighbors_search

        np.intp_t n_jobs

        UnionFind U
        np.intp_t result_edges
        np.ndarray result_value_arr
        np.ndarray result_pairs_arr

        # np.ndarray result_extras1_arr # to extract everything else
        # np.ndarray result_extras2_arr # to extract everything else

    def __init__(self, algorithm, tree, max_neighbors_search=16, metric='euclidean', leaf_size=20, n_jobs=4, is_slow=0, **kwargs):

        self.PRECISION = kwargs.get('double_precision', 0.0000001) # this is only relevant if distances between datapoints are super small
        self.n_jobs = n_jobs

        if algorithm == 0:
            self.dist_tree = tree
            self.tree = KDTree(tree.data, metric=metric, leaf_size=leaf_size, **kwargs)
            self.num_points = self.tree.data.shape[0]
        elif algorithm == 1:
            self.dist_tree = tree
            self.tree = BallTree(tree.data, metric=metric, leaf_size=leaf_size, **kwargs)
            self.num_points = self.tree.data.shape[0]
        elif algorithm == 2:
            self.dist_tree = PairwiseDistanceTreeGeneric(tree.shape[0], tree)
            self.tree = tree
            self.num_points = self.tree.shape[0]
        elif algorithm == 3:
            self.dist_tree = PairwiseDistanceTreeSparse(tree.shape[0], tree)
            self.tree = tree
            self.num_points = self.tree.shape[0]
        else:
            raise ValueError('algorithm value '+str(algorithm)+' is not valid')

        self.max_neighbors_search = max_neighbors_search

        # self.num_features = self.tree.data.shape[1]

        self.U = UnionFind(self.num_points)

        self.result_edges = 0

        self.result_pairs_arr = np.empty((self.num_points*2 - 2))
        self.result_value_arr = np.empty(self.num_points - 1)

        # self.result_extras1_arr = np.empty(self.num_points - 1)
        # self.result_extras2_arr = np.empty(self.num_points - 1)

        self._compute_tree_edges(is_slow)

    cpdef tuple get_tree(self):
        return (self.result_pairs_arr[:self.result_edges*2].astype(int), self.result_value_arr[:self.result_edges])

    cpdef tuple get_extras(self):
        return (self.result_extras1_arr[:self.result_edges], self.result_extras2_arr[:self.result_edges])

    cdef void result_add_edge_debug(self, np.intp_t a, np.intp_t b, Relation* rel):
        cdef np.intp_t i

        i = self.result_edges
        self.result_edges += 1

        self.result_pairs_arr[2*i] = a
        self.result_pairs_arr[2*i + 1] = b
        self.result_value_arr[i] = rel.dia_dis
        # self.result_extras1_arr[i] = rel.rec_rank
        # self.result_extras2_arr[i] = rel.my_members

        # print(self.result_edges, '=================add===============', self.result_value_arr[i], ":", a,b, rel.reciprocity, '=', rel.rec_dis, rel.rec_rank, 'ranks', rel.my_rank, rel.rec_rank, rel.my_scope, 'm',rel.my_members, rel.rec_members)
        # print(rel.reciprocity, self.result_edges, a,b, rel.rec_dis, 'rR', rel.my_rank, rel.rec_rank, 'Dd', rel.rec_dis, rel.my_dis, 'Mm', rel.my_members, rel.rec_members)

    cdef void result_add_edge(self, np.intp_t a, np.intp_t b, np.double_t v):
        cdef np.intp_t i

        i = self.result_edges
        self.result_edges += 1

        self.result_pairs_arr[2*i] = a
        self.result_pairs_arr[2*i + 1] = b
        self.result_value_arr[i] = v


    cdef bint _pure_reciprocity(self, np.intp_t i, np.ndarray[np.intp_t, ndim=2] knn_indices, np.ndarray[np.double_t, ndim=2] knn_dist, Relation* rel, np.intp_t* infinitesimal):
        """Finding pure reciprocal pairs(both ranks = 2)
        And deals with equal objects.
        Runs as initialization, short version of evaluate_reciprocity.
        Fixes problems when amount of same objects are less than K neighbors.
    
        Parameters
        ----------
    
        i : int
            index of the subject
    
        knn_indices: ndarray, shape (n_samples, n_features, )
            Array of arrays. Indices of first K neighbors(including itself, meaning i).
        
        knn_dist: ndarray, shape (n_samples, n_features, )
            Array of arrays. Distances of first K neighbors(including itself, meaning zero).
        
        rel: Relation
            Part of the output. Stores all significant parameters.            
            rel.reciprocity is zero if values are equal - it will lead to relaunch.
            rel.reciprocity is slightly different than in `evaluate_reciprocity`
            
        Infinitesimal: int
            Part of the output. Return distances smaller than the self.PRECISION level.
            Used to inform the user about hidden parameter.
            
        Returns
        -------
        res : bint
            Success if not zero.
        """
        cdef:
            np.intp_t ranki, j, \
                rank_left, rank_right, \
                parent

            np.double_t dis

        parent = self.U.fast_find(i)
        indices, distances = knn_indices[i], knn_dist[i]
        for ranki in range(0, self.max_neighbors_search + 1):
            j = indices[ranki]
            if parent == self.U.fast_find(j):
                continue

            dis = distances[ranki]
            if dis == 0.: # degenerate case.
                rel.reciprocity = 0.
                rel.target = j
                # rel.my_rank = ranki
                # rel.rec_rank = ranki
                # rel.my_dis = 0.
                rel.dia_dis = 0.
                # rel.my_members = ranki
                # rel.rec_members = 1
                # rel.index = i

                return ranki + 1

            if dis <= self.PRECISION:
                infinitesimal += 1

            rank_left = bisect.bisect(distances, dis + self.PRECISION)
            if rank_left > 2:
                return 0

            rank_right = bisect.bisect(knn_dist[j], dis + self.PRECISION)
            if rank_right > 2:
                return 0

            rel.reciprocity = dis
            rel.target = j
            # rel.my_rank = 2
            # rel.rec_rank = 2
            # rel.my_dis = dis
            rel.dia_dis = dis
            # rel.my_members = 1
            # rel.rec_members = 1
            # rel.index = i

            return 2
        return 0

    cdef bint _evaluate_reciprocity_slow(self, np.intp_t i, np.ndarray[np.intp_t, ndim=2] knn_indices, np.ndarray[np.double_t, ndim=2] knn_dist, Relation *rel):

        cdef:
            np.intp_t ranki, j, \
                parent, \
                rank_left, rank_right, \
                res = 0

            np.double_t best, order, \
                dis, rank_dis

            np.ndarray indices, ind_opp
            np.ndarray distances, dis_opp

        parent = self.U.fast_find(i)
        indices, distances = knn_indices[i], knn_dist[i]

        best = INF
        # print(self.max_neighbors_search, 'yo',distances)
        for ranki in range(0, self.max_neighbors_search + 1):
            j = indices[ranki]
            dis = distances[ranki]

            if parent == self.U.fast_find(j):
                continue

            if dis >= best + self.PRECISION:
                break

            dis_opp = knn_dist[j]
            rank_right = bisect.bisect(dis_opp, dis + self.PRECISION) - 1 # !reminder! bisect.bisect(dis_opp, dis) >= bisect.bisect_left(dis_opp, dis)
            if ranki > rank_right:
                continue

            rank_left = bisect.bisect(distances, dis + self.PRECISION) - 1
            if rank_left > rank_right:
                continue

            rank_dis = distances[rank_right]
            rank_left = len(set(indices[:rank_left+1]) - set(knn_indices[j][:rank_right+1])) + 1

            order = 1. * rank_right/rank_left * rank_dis

            # print(i, order < best, order, '=', rank_right, m, rank_dis, 'rl', rank_left, dis)
            # print(indices[:rank_left+1], knn_indices[j][:rank_right+1])
            # print(distances[:rank_left+1], knn_dist[j][:rank_right+1])
            # print(distances[:rank_left+2], knn_dist[j][:rank_right+2])

            if order < best: # minimizing
                res = 1
                best = order

                rel.reciprocity = best
                rel.target = j
                # rel.my_dis = dis
                rel.dia_dis = rank_dis
                # rel.loop_rank = ranki
                # rel.my_rank = rank_left
                # rel.rec_rank = rank_right
                # rel.my_scope = scope_left
                # rel.my_members = members
            # print (i,j, rel.reciprocity, ':', rel.my_dis, rel.rec_dis, ":", rel.my_rank, rel.rec_rank, rel.my_scope, ".", rel.my_members)
            #     print (i,j, rel.reciprocity, ':', rel.my_dis, rel.rec_dis, ":", rel.my_rank, rel.rec_rank, rel.my_scope, ".", rel.my_members)
                # rel.value = pow(rel.rec_dis, 2) * (rel.rec_rank) * pow(1.*rel.my_members/rel.rec_members, 0.5)

        return res

    cdef bint _evaluate_reciprocity_fast(self, np.intp_t i, np.ndarray[np.intp_t, ndim=2] knn_indices, np.ndarray[np.double_t, ndim=2] knn_dist, Relation *rel):

        cdef:
            np.intp_t ranki, j, \
                parent, \
                rank_left, rank_right, \
                res = 0

            np.double_t best, order, \
                dis, rank_dis

            np.ndarray indices, ind_opp
            np.ndarray distances, dis_opp

        parent = self.U.fast_find(i)
        indices, distances = knn_indices[i], knn_dist[i]

        best = INF
        # print(self.max_neighbors_search, 'yo',distances)
        for ranki in range(0, self.max_neighbors_search + 1):
            j = indices[ranki]
            dis = distances[ranki]

            if parent == self.U.fast_find(j):
                continue

            if dis >= best + self.PRECISION:
                break

            dis_opp = knn_dist[j]
            rank_right = bisect.bisect(dis_opp, dis + self.PRECISION) - 1 # !reminder! bisect.bisect(dis_opp, dis) >= bisect.bisect_left(dis_opp, dis)
            if ranki > rank_right:
                continue

            rank_left = bisect.bisect(distances, dis + self.PRECISION) - 1
            if rank_left > rank_right:
                continue

            rank_dis = distances[rank_right]
            # rank_left = 1.*len(set(indices[:rank_left+1]) - set(knn_indices[j][:rank_right+1])) + 1 # slow/proper
            order = 1. * rank_right/rank_left * rank_dis

            # print(i, order < best, order, '=', rank_right, m, rank_dis, 'rl', rank_left, dis)
            # print(indices[:rank_left+1], knn_indices[j][:rank_right+1])
            # print(distances[:rank_left+1], knn_dist[j][:rank_right+1])
            # print(distances[:rank_left+2], knn_dist[j][:rank_right+2])

            if order < best: # minimizing
                res = 1
                best = order

                rel.reciprocity = best
                rel.target = j
                # rel.my_dis = dis
                rel.dia_dis = rank_dis
                # rel.loop_rank = ranki
                # rel.my_rank = rank_left
                # rel.rec_rank = rank_right
                # rel.my_scope = scope_left
                # rel.my_members = members
            # print (i,j, rel.reciprocity, ':', rel.my_dis, rel.rec_dis, ":", rel.my_rank, rel.rec_rank, rel.my_scope, ".", rel.my_members)
            #     print (i,j, rel.reciprocity, ':', rel.my_dis, rel.rec_dis, ":", rel.my_rank, rel.rec_rank, rel.my_scope, ".", rel.my_members)
                # rel.value = pow(rel.rec_dis, 2) * (rel.rec_rank) * pow(1.*rel.my_members/rel.rec_members, 0.5)

        return res


    cdef _compute_tree_edges(self, is_slow):
        # DRUHG
        # computes DRUHG Spanning Tree
        # uses heap

        cdef:
            np.intp_t i,j, \
                warn, infinitesimal
            np.double_t v

            list heap
            Relation rel

            np.ndarray[np.double_t, ndim=2] knn_dist
            np.ndarray[np.intp_t, ndim=2] knn_indices

        if self.tree.data.shape[0] > 16384 and self.n_jobs > 1: # multicore 2-3x speed up for big datasets
            split_cnt = self.num_points // self.n_jobs
            datasets = []
            for i in range(self.n_jobs):
                if i == self.n_jobs - 1:
                    datasets.append(np.asarray(self.tree.data[i*split_cnt:]))
                else:
                    datasets.append(np.asarray(self.tree.data[i*split_cnt:(i+1)*split_cnt]))

            knn_data = Parallel(n_jobs=self.n_jobs)(
                delayed(self.tree.query)
                (points,
                 self.max_neighbors_search + 1,
                 dualtree=True,
                 breadth_first=True
                 )
                for points in datasets)
            knn_dist = np.vstack([x[0] for x in knn_data])
            knn_indices = np.vstack([x[1] for x in knn_data])
        else:
            knn_dist, knn_indices = self.dist_tree.query(
                        self.tree.data,
                        k=self.max_neighbors_search + 1,
                        dualtree=True,
                        breadth_first=True,
                        )

        warn, infinitesimal = 0, 0
        heap = []
#### Initialization of pure reciprocity then ranks are less than 2
        i = self.num_points
        while i:
            i -= 1
            if knn_dist[i][0] < 0.:
                print ('Distances cannot be negative! Exiting. ', i, knn_dist[i][0])
                return

            rel.reciprocity = 1.
            if self._pure_reciprocity(i, knn_indices, knn_dist, &rel, &infinitesimal):
                self.U.union(i, rel.target)
                self.result_add_edge(i, rel.target, rel.dia_dis)

            if rel.reciprocity == 0.: # values match
                warn += 1
                i += 1 # need to relaunch same values
                continue

                # print ('pure', i,j)
#### initialization of reciprocities
            if (is_slow and self._evaluate_reciprocity_slow(i, knn_indices, knn_dist, &rel))\
            or (not is_slow and self._evaluate_reciprocity_fast(i, knn_indices, knn_dist, &rel)):
                heapq.heappush(heap, (rel.reciprocity, i, rel.target, rel.dia_dis))
                # print('RESULT', rel.reciprocity, i)
        # return
        if warn > 0:
            print ('A lot of values(',warn,') are the same. Try increasing max_neighbors_search(',self.max_neighbors_search,') parameter.')

        if infinitesimal > 0:
            print ('Some distances(', infinitesimal, ') are smaller than self.PRECISION (', self.PRECISION,') level. Try decreasing double_precision parameter.')

        if self.result_edges >= self.num_points - 1:
            print ('Two subjects only')
            return

        heapq.heappush(heap, (INF, -1,-1,0.))
        if is_slow:
            while self.result_edges < self.num_points - 1:
                best, i, j, v = heapq.heappop(heap)

                if best == INF:
                    print (str(self.num_points - 1 - self.result_edges) +' not connected edges. It is a forest. Try increasing max_neighbors(max_ranking) value '+str(self.max_neighbors_search)+' for a better result.')
                    break

                if self.U.fast_find(i)!=self.U.fast_find(j):
                    self.U.union(i, j)
                    self.result_add_edge(i, j, v)
                if self._evaluate_reciprocity_slow(i, knn_indices, knn_dist, &rel):
                    heapq.heappush(heap, (rel.reciprocity, i, rel.target, rel.dia_dis)) # updated value is in
        else:
            while self.result_edges < self.num_points - 1:
                best, i, j, v = heapq.heappop(heap)

                if best == INF:
                    print (str(self.num_points - 1 - self.result_edges) +' not connected edges. It is a forest. Try increasing max_neighbors(max_ranking) value '+str(self.max_neighbors_search)+' for a better result.')
                    break

                if self.U.fast_find(i)!=self.U.fast_find(j):# по идее можем присоединять? но тогда нужно тащить target и diadis
                    self.U.union(i, j)
                    self.result_add_edge(i, j, v)
                if self._evaluate_reciprocity_fast(i, knn_indices, knn_dist, &rel):
                    heapq.heappush(heap, (rel.reciprocity, i, rel.target, rel.dia_dis)) # updated value is in

        return
        # less storage, reevaluating to get the data
        # heapq.heappush(heap, (INF, -1))
        # if is_slow:
        #     while self.result_edges < self.num_points - 1:
        #         best, i = heapq.heappop(heap)
        #         if best == INF:
        #             print (str(self.num_points - 1 - self.result_edges) +' not connected edges. It is a forest. Try increasing max_neighbors(max_ranking) value '+str(self.max_neighbors_search)+' for a better result.')
        #             break
        #         if self._evaluate_reciprocity_slow(i, knn_indices, knn_dist, &rel):
        #             if rel.reciprocity <= best + self.PRECISION:
        #                 self.U.union(i, rel.target)
        #                 self.result_add_edge(i, rel.target, &rel)
        #                 if self._evaluate_reciprocity_slow(i, knn_indices, knn_dist, &rel):
        #                     heapq.heappush(heap, (rel.reciprocity, i)) # updated value is in
        #             else: # the result got worse. When equal values are connected.
        #                 heapq.heappush(heap, (rel.reciprocity, i))
        # else:
        #     while self.result_edges < self.num_points - 1:
        #         best, i = heapq.heappop(heap)
        #         if best == INF:
        #             print (str(self.num_points - 1 - self.result_edges) +' not connected edges. It is a forest. Try increasing max_neighbors(max_ranking) value '+str(self.max_neighbors_search)+' for a better result.')
        #             break
        #         # по идее можем присоединять? но тогда нужно тащить target и diadis
        #         if self._evaluate_reciprocity_fast(i, knn_indices, knn_dist, &rel):
        #             if rel.reciprocity <= best + self.PRECISION:
        #                 self.U.union(i, rel.target)
        #                 self.result_add_edge(i, rel.target, &rel)
        #                 if self._evaluate_reciprocity_fast(i, knn_indices, knn_dist, &rel):
        #                     heapq.heappush(heap, (rel.reciprocity, i)) # updated value is in
        #             else: # the result got worse. When equal values are connected.
        #                 heapq.heappush(heap, (rel.reciprocity, i))



# cython: language_level=3
# cython: boundscheck=False
# cython: nonecheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True

import numpy as np
cimport numpy as np

cdef class Amalgamation (object):
    cdef:
        np.intp_t size
        np.intp_t clusters
        dict energies

        cdef Amalgamation merge_amalgamations(self, np.double_t g, Amalgamation other, np.double_t jump1, np.double_t jump2)
        np.double_t border_overcoming(self, np.double_t g, Amalgamation other, np.double_t PRECISION)
        np.double_t border_overcoming_rev(self, np.double_t g, Amalgamation other, np.double_t PRECISION)

        void _amalgamate(self, np.intp_t size, np.intp_t clusters, dict energies)



# cython: language_level=3
# cython: boundscheck=False
# cython: nonecheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True

# Labels nodes given mst-edgepairs and cluster-parents of DRUHG's results.
# Also provides tools for label manipulations, such as:
# * Treats small clusters as outliers on-demand
# * Breaks big clusters on-demand
# * Glues outliers to the nearest clusters on-demand
#
# Author: Pavel "DRUHG" Artamonov
# License: 3-clause BSD

import numpy as np
cimport numpy as np

from ._druhg_amalgamation import Amalgamation
from ._druhg_amalgamation cimport Amalgamation

cdef class UnionFind (object):

    cdef:
        np.ndarray parent_arr
        np.intp_t *parent

        np.intp_t next_label
        np.intp_t full_size

    def __init__(self, N):
        self.full_size = N
        self.next_label = N + 1

        self.parent_arr = np.zeros(2 * N, dtype=np.intp)
        self.parent = (<np.intp_t *> self.parent_arr.data)

        # self._fill_structure(edges_arr, num_edges)

    cdef np.intp_t has_parent(self, np.intp_t n):
        return self.parent[n]

    cdef np.intp_t fast_find(self, np.intp_t n):
        cdef np.intp_t p, temp

        p = self.parent[n]
        if p == 0:
            return n
        while self.parent[p] != 0:
            p = self.parent[p]

        # label up to the root
        while p != n:
            temp = self.parent[n]
            self.parent[n] = p
            n = temp

        return p

    cdef np.intp_t passive_find(self, np.intp_t n):
        cdef np.intp_t p, temp

        p = self.parent[n]
        if p == 0:
            return n
        while self.parent[p] != 0:
            p = self.parent[p]

        return p

    cdef np.intp_t passive_union(self, np.intp_t aa, np.intp_t bb):
        # aa, bb = self.fast_find(aa), self.fast_find(bb)
        cdef np.intp_t ret

        ret = self.next_label
        self.parent[aa] = self.parent[bb] = ret
        self.next_label += 1
        return ret

    cdef list discard_clusters(self, np.intp_t n, set clusters):
        cdef:
            np.intp_t p, temp
            list ret

        ret = []
        p = self.parent[n]
        if p == 0:
            return ret
        while self.parent[p] != 0:
            if p in clusters:
                ret.append(p)
            p = self.parent[p]
        if ret:
            ret.pop()
        return ret

    cdef np.intp_t label_find(self, np.intp_t n, set clusters):
        cdef np.intp_t p, temp, label

        label = -1

        p = self.parent[n]
        if p == 0:
            return label

        if p in clusters:
            label = p
        while self.parent[p] != 0:
            p = self.parent[p]
            if p in clusters:
                label = p
        if label >= 0:
            while label != n:
                temp = self.parent[n]
                self.parent[n] = label
                n = temp
        return label

cdef void fixem(np.ndarray edges_arr, np.intp_t num_edges, np.ndarray result):
    cdef:
        np.intp_t p, a, b, dontstop
        set new_results, links
        list new_path, restart

    new_results = set()
    new_path = []
    restart = []
    for p in range(0, num_edges):
        a, b = edges_arr[2*p], edges_arr[2*p + 1]
        if result[a] < 0 and result[b] < 0:
            new_results.update([a,b])
            new_path.append((a,b))
            continue
        elif result[b] < 0:
            a,b = b,a
        elif result[a] >= 0:
            continue
        res = result[b]
        result[a] = res
        if a in new_results:
            links = set([a])
            dontstop = 1
            while dontstop:
                dontstop = 0
                for path in list(new_path):
                    a, b = path
                    if a in links or b in links:
                        result[a] = result[b] = res
                        # print ('new_r', a, b, res)
                        links.update([a,b])
                        new_path.remove(path)
                        dontstop = 1

    return

def emerge_clusters(UnionFind U, np.ndarray edges_arr, np.ndarray values_arr, np.intp_t limit1, np.intp_t limit2, \
                    list exclude, is_reciprocal = 1, **kwargs):

    cdef:
        np.intp_t e1,e2,e3, p1,p2, i, c
        np.double_t v

        Amalgamation being, being1, being2, being3
        np.double_t jump1, jump2
        list disc
        dict d
        set clusters
        np.double_t PRECISION

    PRECISION = kwargs.get('double_precision2', kwargs.get('double_precision', 0.0000001)) # this is only relevant if distances between datapoints are super small
    # extras = kwargs.get('extras_arr', None)

    being = Amalgamation()
    being1 = being
    being2 = being

    d = {}
    clusters = set()

    for i, v in enumerate(values_arr):
        e1, e2 = edges_arr[2*i], edges_arr[2*i+1]
        # print ('edges', e1, e2)
        p1, p2 = U.passive_find(e1), U.passive_find(e2)
        being1, being2 = being, being

        if U.has_parent(e1):
            being1 = d.pop(p1)
        if U.has_parent(e2):
            being2 = d.pop(p2)

        if is_reciprocal:
            jump1 = being1.border_overcoming_rev(v, being2, PRECISION)
            jump2 = being2.border_overcoming_rev(v, being1, PRECISION)
        else:
            jump1 = being1.border_overcoming(v, being2, PRECISION)
            jump2 = being2.border_overcoming(v, being1, PRECISION)

# TODO: добавить ворнинг на пресижн?

        if being1.size > 1 \
            and jump1 >= 0 \
            and limit1 < being1.size < limit2 \
            and p1 not in exclude:
            clusters.add(p1)

        if being2.size > 1 \
            and jump2 >= 0 \
            and limit1 < being2.size < limit2 \
            and p2 not in exclude:
            clusters.add(p2)

        being3 = being1.merge_amalgamations(v, being2, jump1, jump2)

        e3 = U.passive_union(U.passive_find(e1), U.passive_find(e2))
        # print(p1,p2,e3)
        d[e3] = being3
        if being2 != being:
            del being2

        disc = U.discard_clusters(e1, clusters)
        for c in disc:
            clusters.discard(c)
        disc = U.discard_clusters(e2, clusters)
        for c in disc:
            clusters.discard(c)

        U.label_find(e1, clusters)
        U.label_find(e2, clusters)

    return clusters


cpdef np.ndarray label(np.ndarray edges_arr, np.ndarray values_arr, int size = 0, list exclude = None, \
                       np.intp_t limit1 = 0, np.intp_t limit2 = 0, np.intp_t fix_outliers = 1, is_reciprocal = 1):
    """Returns cluster labels.
    
    Uses the results of DRUHG MST-tree algorithm(edges and values).
    Marks data-points with corresponding parent index of a cluster.
    Exclude list breaks passed clusters by their parent index.
    The parameters `limit1` and 'limit2' allows the clustering to declare noise points.
    Outliers-noise marked by -1.

    Parameters
    ----------
    edges_arr : ndarray
        Edgepair nodes of mst.
    
    values_arr : ndarray
        Edge values.
        
    size : int
        Amount of nodes.
        
    exclude : list
        Clusters with parent-index from this list will not be formed. 
    
    limit1 : int, optional (default=sqrt(size))
        Clusters that are smaller than this limit treated as noise. 
        Use 1 to find True outliers.
        
    limit2 : int, optional (default=size)
        Clusters with size OVER this limit treated as noise. 
        Use it to break down big clusters.
 
    fix_outliers: int, optional (default=0)
        All outliers will be assigned to the nearest cluster.

    is_reciprocal: int, optional (default=1)
        The formula turns upside down.

    Returns
    -------

    labels : array [size]
       An array of cluster labels, one per data-point. Unclustered points get
       the label -1.
    """

    cdef:
        set clusters
        int i

        np.ndarray result_arr
        np.intp_t *result


    if limit1 <= 0:
        limit1 = int(np.ceil(np.sqrt(size)))
        print ('label: default value for limit1 is used, clusters below '+str(limit1)+' are considered as noise.')

    if limit2 <= 0:
        limit2 = size
        print ('label: default value for limit2 is used, clusters above '+str(limit2)+' will not be formed.')

    if size == 0:
        size = int(len(edges_arr)/2 + 1)

    if not exclude:
        exclude = []

    result_arr = -1*np.ones(size, dtype=np.intp)
    result = (<np.intp_t *> result_arr.data)

    U = UnionFind(size)
    clusters = emerge_clusters(U, edges_arr, values_arr, limit1, limit2, exclude, is_reciprocal)

    i = size
    while i:
        i -= 1
        result[i] = U.label_find(i, clusters)

    if fix_outliers != 0 and len(np.unique(result_arr))>1:
        fixem(edges_arr, len(values_arr), result_arr)

    return result_arr


cdef np.ndarray pretty(np.ndarray labels_arr):
    """ Relabels to pretty positive integers. 
    """
    cdef np.intp_t i, p, label, max_label
    cdef np.ndarray[np.intp_t, ndim=1] result_arr
    cdef dict converter
    cdef np.intp_t* result

    result_arr = -1*np.ones(len(labels_arr), dtype=np.intp)
    result = (<np.intp_t *> result_arr.data)

    converter = {-1: -1}
    max_label = 0
    i = len(labels_arr)
    while i:
        i -= 1
        p = labels_arr[i]
        if p in converter:
            label = converter[p]
        else:
            label = max_label
            converter[p] = max_label
            max_label += 1
        result[i] = label

    return result_arr



# cython: language_level=3
# cython: boundscheck=False
# cython: nonecheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True

# amalgamation structure that can become a cluster
# Author: Pavel "DRUHG" Artamonov
# License: 3-clause BSD

import numpy as np
cimport numpy as np
import sys

from libc.math cimport fabs, pow

cdef np.double_t merge_means(np.intp_t na, np.double_t meana,
                             np.intp_t nb, np.double_t meanb
                            ):
    # https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
    # Chan et al.[10] Welford's online algorithm
    cdef np.double_t delta

    # nx = na + nb
    delta = meanb - meana
    delta = meana + delta*nb/(na + nb)
    # use this for big n's
    # mu = (mu*n + mu_2*n_2) / nx
    # m2a = m2a + m2b + delta**2*na*nb/nx
    return delta

cdef class Amalgamation (object):
    # declarations are in pxd file
    # https://cython.readthedocs.io/en/latest/src/userguide/sharing_declarations.html

    def __init__(self, int size = 1, int clusters = 1):
        self.size = size
        self.clusters = clusters
        self.energies = None

    cdef void _amalgamate(self, np.intp_t size, np.intp_t clusters, dict o_energies):
        self.size += size
        self.clusters += clusters

        energy = dict()
        for elem in set(self.energies) | set(o_energies):
            v1 = self.energies.get(elem, (0,0.,))
            v2 = o_energies.get(elem, (0,0.,))
            energy[elem] = (v1[0]+v2[0], v1[1]+v2[1],)

        self.energies = energy

    cdef Amalgamation merge_amalgamations(self, np.double_t g, Amalgamation other, np.double_t jump1, np.double_t jump2):
        cdef:
            np.intp_t osize, oclusters
            Amalgamation ret
        ret = self
        if self.size == 1:
            ret = Amalgamation() # the self will be reused
# ----------------------
        osize, oclusters, o_energies = other.size, other.clusters, other.energies
# ----------------------
        if jump1 >= 0:
            ret.clusters = 1
            ret.energies = {ret.size : (1., jump1,)}
# ----------------------
        if jump2 >= 0:
            oclusters = 1
            o_energies = {osize : (1., jump2,)}
# ----------------------

        ret._amalgamate(osize, oclusters, o_energies)
        return ret


    cdef np.double_t border_overcoming(self, np.double_t g, Amalgamation other, np.double_t PRECISION):
        # returns negative if cluster didn't form
        # limit_to_ought:
        # Die Schranke und das Sollen
        # can a new whole overcome its' parts?
        cdef np.double_t insides, limit
        cdef np.double_t clusters, osize
        cdef np.double_t n, c, e

        if self.energies is None:
            return g # first connection always clusterize

        clusters = self.clusters - 1
        osize = other.size

        insides, limit = 0., 0.
        for n in self.energies:
            c,e = self.energies[n]
            insides += n/min(n, clusters) * e
            limit += min(n, osize)/n * c
            # print (n, e, c, n/min(n, clusters) * e, min(n, osize)/n * c)

        limit *= g-PRECISION

        # print('..',g, 'c', clusters, other.clusters, 's', self.size, other.size, 'r', insides,  limit)
        # print(list(self.energies))
        # print (insides > limit,'=============================================')

        if insides < limit:
            return g

        return -1.

    cdef np.double_t border_overcoming_rev(self, np.double_t g, Amalgamation other, np.double_t PRECISION):
        # returns negative if cluster didn't form
        # limit_to_ought:
        # Die Schranke und das Sollen
        # can a new whole overcome its' parts?
        cdef np.double_t insides, limit
        cdef np.double_t clusters, osize
        cdef np.double_t n, c, e

        if self.energies is None:
            return pow(g, -1) # first connection always clusterize

        clusters = self.clusters - 1
        osize = other.size

        insides, limit = 0., 0.
        for n in self.energies:
            c,e = self.energies[n]
            insides += min(n, clusters)/n * e
            limit += n/min(n, osize) * c

        insides *= g-PRECISION
        if insides > limit:
            return pow(g, -1)

        return -1.
